import socket

from typing import Literal, overload
from collections.abc import Callable


class SackMessage:
    type: Literal["CONNECT", "DISCONNECT", "TEXT", "GETNICKNAMES"]
    username: str
    text: str | None

    @overload
    def __init__(self, type: Literal["GETNICKNAMES"], username: str, text: str | None = None): ...
    @overload
    def __init__(self, type: Literal["CONNECT", "DISCONNECT"], username: str) -> None: ...
    @overload
    def __init__(self, type: Literal["TEXT"], username: str, text: str) -> None: ...

    def __init__(self, type, username, text=None) -> None:
        self.type = type
        self.username = username
        self.text = text

    def to_bytes(self) -> bytes:
        message = f"{self.type}\n{self.username}".encode()
        message = len(message).to_bytes(1, "big") + message
        if self.text is None:
            return message
        assert self.text is not None
        text = self.text.encode()
        message += b"\n" + len(text).to_bytes(2, "big") + text
        return message


def receive_message(socket: socket.socket, on_empty: Callable) -> SackMessage | None:
    message_len = socket.recv(1)  # ConnectionResetError
    if not message_len:
        return on_empty()
    message_len = int.from_bytes(message_len, "big")
    raw_message = socket.recv(message_len)
    raw_message = raw_message.decode()
    message_parts = raw_message.split("\n")
    if not len(message_parts) == 2:
        return None
    type = message_parts[0]
    username = message_parts[1]
    if type not in ("CONNECT", "TEXT", "DISCONNECT", "GETNICKNAMES"):
        return None
    if type != "TEXT":
        return SackMessage(type, username)
    sep = socket.recv(1)
    if sep != b"\n":
        return None
    text_len = socket.recv(2)
    text_len = int.from_bytes(text_len, "big")
    text = socket.recv(text_len).decode()
    return SackMessage(type, username, text)



import os
import queue
import socket
import logging
import selectors
import threading

from typing import cast
from dataclasses import dataclass
from collections.abc import Callable

from sack.models.protocol import SackMessage, receive_message


log = logging.getLogger("server")
blog = logging.getLogger("broadcaster")


@dataclass
class ClientData:
    username: str | None = None

    @property
    def is_registered(self) -> bool:
        return self.username is not None


class SackServer:
    def __init__(self, host: str, port: int) -> None:
        self.host = host
        self.port = port

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((host, port))
        self._socket = s

        sock_read, sock_write = socket.socketpair()
        self._STOP = sock_read
        self._stop_controller = sock_write

        self._registry = selectors.DefaultSelector()
        self._registry.register(self._socket, selectors.EVENT_READ)
        self._registry.register(self._STOP, selectors.EVENT_READ)

        def get_connections() -> list[socket.socket]:
            return cast(list[socket.socket], [key.fileobj for key in self._get_client_keys()])

        self._broadcaster = Broadcaster(connections_getter=get_connections)

    def serve(self):
        self._socket.setblocking(False)
        self._socket.listen()
        self._broadcaster.run()

        log.info("Started at %s:%d", self.host, self.port)
        log.debug("PID: %d", os.getpid())

        while True:
            events = self._registry.select()
            for key, mask in events:
                assert isinstance(key.fileobj, socket.socket)

                if key.fileobj is self._socket:
                    self._accept_connection()

                elif key.fileobj is self._STOP:
                    self._STOP.recv(1)
                    log.info("stopping server")
                    return

                else:
                    assert mask == selectors.EVENT_READ
                    message = self._receive_client_message(key.fileobj)
                    if message is None:
                        continue
                    log.info("received message of type %s", message.type)
                    client_data: ClientData = key.data
                    if message.type != "CONNECT" and not client_data.is_registered:
                        continue
                    if message.type == "DISCONNECT":
                        message.username = key.data.username
                        log.info("client disconnects")
                        self._unregister(key.fileobj)
                    if message.type == "CONNECT":
                        username = message.username
                        if username in self._get_usernames():
                            key.fileobj.sendall(b"NO")
                            continue
                        else:
                            key.fileobj.sendall(b"OK")
                            key.data.username = message.username
                    if message.type == "GETNICKNAMES":
                        ret = "\n".join(self._get_usernames())
                        msg = SackMessage("GETNICKNAMES", key.data.username, ret)
                        key.fileobj.sendall(msg.to_bytes())
                        continue

                    self._broadcaster.broadcast(message.to_bytes())

    def stop(self):
        self._stop_controller.send(b"\0")

    def _accept_connection(self):
        conn, addr = self._socket.accept()
        conn.setblocking(False)
        self._registry.register(conn, selectors.EVENT_READ, ClientData())
        log.info("Accepted connection from %s", addr)

    def _unregister(self, sock: socket.socket):
        sock.close()
        self._registry.unregister(sock)

    def _receive_client_message(self, socket: socket.socket) -> SackMessage | None:
        def on_empty():
            return SackMessage("DISCONNECT", "")

        return receive_message(socket, on_empty)

    def _get_usernames(self) -> list[str]:
        return [username for key in self._get_client_keys() if (username := key.data.username)]

    def _get_client_keys(self):
        for key in self._registry.get_map().values():
            if isinstance(key.data, ClientData):
                yield key

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self._broadcaster.shutdown()
        self._socket.close()
        self._registry.close()


class Broadcaster:
    def __init__(self, connections_getter: Callable[[], list[socket.socket]]) -> None:
        self._msg_queue = queue.Queue()
        self._worker = threading.Thread(target=self._broadcast_worker)
        self._STOP = object()
        self._get_connections = connections_getter

    def run(self):
        self._worker.start()

    def shutdown(self):
        self._msg_queue.put(self._STOP)

    def broadcast(self, message: bytes):
        self._msg_queue.put(message)

    def _broadcast_worker(self):
        while True:
            msg = self._msg_queue.get()
            if msg is self._STOP:
                self._msg_queue.task_done()
                break
            connections = self._get_connections()
            blog.info("broadcasting message to %d clients", len(connections))
            for conn in connections:
                conn.sendall(msg)
            self._msg_queue.task_done()



import socket
import multiprocessing

from typing import TYPE_CHECKING
from contextlib import suppress

from textual import on
from textual.app import ComposeResult
from textual.screen import Screen, ModalScreen
from textual.binding import Binding
from textual.message import Message
from textual.widgets import Input, Label, Button, Static, ContentSwitcher
from textual.containers import Center, Container, VerticalScroll, HorizontalGroup

from sack.util import ColorsManager, get_sidebar_user, get_common_footer, get_id_from_color, make_keybinding_text
from sack.assets import CHAT_HELP, JOIN_HELP, SACK_ABOUT, SERVER_HELP, WELCOME_HELP
from sack.models import SackServer, SackMessage, AsyncSackClient, SackClientServerError, SackClientUsernameError
from sack.components import (
    Option,
    HelpTab,
    Options,
    FormField,
    TextInput,
    ChatHeader,
    FormButton,
    FormErrors,
    ChatMessage,
    ChatSidebar,
    HelpKeybinding,
    VimVerticalScroll,
)
from sack.keybindings import CHAT_KB, HELP_KB, ABOUT_KB, FORMS_KB, WELCOME_KB


if TYPE_CHECKING:
    from sack.main import SackApp


COMMON_BINDINGS = [
    Binding("escape", "app.pop_screen"),
    Binding("tab, ctrl+j", "app.focus_next", priority=True),
    Binding("shift+tab, ctrl+k", "app.focus_previous", priority=True),
]


class ServerPromptScreen(Screen):
    app: "SackApp"

    BINDINGS = COMMON_BINDINGS

    def compose(self) -> ComposeResult:
        yield from self.app.get_header()
        with Container(id="form", classes="container"):
            yield Label("Create server", classes="form-title")
            yield FormErrors()
            yield FormField("port", "Port:", type="integer", max_length=5)
            yield FormButton("Next")
        yield from get_common_footer()

    def on_input_changed(self, _) -> None:
        form_error = self.query_one(FormErrors)
        form_error.reset()

    async def on_button_pressed(self, _):
        await self.app.cleanup()
        port = self.query_one("#port", Input).value
        form_error = self.query_one(FormErrors)
        if not port:
            form_error.set_error(1, "Port is required")
            return
        port = int(port)
        host = socket.gethostname()
        assert isinstance(host, str)

        event = multiprocessing.Event()

        def server_launcher():
            try:
                with SackServer("0.0.0.0", port) as server:
                    server.serve()
            except Exception:
                event.set()

        server_process = multiprocessing.Process(target=server_launcher, daemon=True)
        self.app.server_process = server_process
        server_process.start()
        if event.wait(0.1):
            form_error.set_error(2, "Port not available")
            return

        form_error.reset()
        client = AsyncSackClient(host=host, port=port)
        await client.connect()
        self.app.client = client
        self.app.push_screen(NicknamePromtScreen(form_title="Create server", button_label="Create"))


class ClientPromptScreen(Screen):
    app: "SackApp"

    BINDINGS = COMMON_BINDINGS

    def compose(self) -> ComposeResult:
        yield from self.app.get_header()
        with Container(id="form", classes="container"):
            yield Label("Join server", classes="form-title")
            yield FormErrors()
            yield FormField("host", "Host:")
            yield FormField("port", "Port:", type="integer", max_length=5)
            yield FormButton("Next")
        yield from get_common_footer()

    def on_input_changed(self, e: Input.Changed) -> None:
        form_error = self.query_one(FormErrors)
        form_error.clear_error(3)
        match e.input.id:
            case "host":
                form_error.clear_error(1)
            case "port":
                form_error.clear_error(2)

    async def on_button_pressed(self, _):
        await self.app.cleanup()
        host = self.query_one("#host", Input).value
        port = self.query_one("#port", Input).value
        form_error = self.query_one(FormErrors)
        if not host:
            form_error.set_error(1, "Host is required")
        if not port:
            form_error.set_error(2, "Port is required")
        if form_error.has_errors(1, 2):
            return
        port = int(port)
        assert isinstance(host, str)

        client = AsyncSackClient(host=host, port=port)

        try:
            await client.connect(timeout=0.1)
        except SackClientServerError:
            form_error.set_error(3, "Server not found")
            return

        form_error.reset()
        self.app.client = client
        self.app.push_screen(NicknamePromtScreen(form_title="Join server", button_label="Join"))


class NicknamePromtScreen(Screen):
    app: "SackApp"

    BINDINGS = COMMON_BINDINGS

    def __init__(self, form_title: str, button_label: str) -> None:
        super().__init__()
        self.form_title = form_title
        self.button_label = button_label

    def on_input_changed(self, _) -> None:
        form_error = self.query_one(FormErrors)
        form_error.reset()

    def compose(self) -> ComposeResult:
        yield from self.app.get_header()
        with Container(id="form", classes="container"):
            yield Label(self.form_title, classes="form-title")
            yield FormErrors()
            yield FormField("nickname", "Nickname:", max_length=15)
            yield FormButton(self.button_label)
        yield from get_common_footer()

    async def on_button_pressed(self, _):
        nickname = self.query_one("#nickname", Input).value
        form_error = self.query_one(FormErrors)
        if not nickname:
            form_error.set_error(1, "Nickname is required")
            return

        client = self.app.client
        assert client
        client.username = nickname
        try:
            await client.join_request()
        except SackClientUsernameError:
            form_error.set_error(2, "Nickname already taken")
            return

        form_error.reset()
        self.app.push_screen(ChatScreen())


class ChatScreen(Screen):
    app: "SackApp"

    BINDINGS = [
        Binding("enter", "send", priority=True),
        Binding("ctrl+c", "quit", priority=True),
        Binding("f1", "show_help", priority=True),
        Binding("escape", "open_menu"),
        Binding("ctrl+j", "app.focus_next", priority=True),
        Binding("ctrl+k", "app.focus_previous", priority=True),
        Binding("ctrl+s", "toggle_sidebar"),
    ]

    def action_toggle_sidebar(self) -> None:
        self.query_one(ChatSidebar).toggle_class("hidden")

    class MessageReceived(Message):
        def __init__(self, msg: SackMessage) -> None:
            super().__init__()
            self.msg = msg

    class ServerDown(Message):
        pass

    def __init__(self) -> None:
        super().__init__()
        assert self.app.client
        self.client = self.app.client
        self.username = self.app.client.username
        self.colors_manager = ColorsManager()

    def compose(self) -> ComposeResult:
        yield ChatSidebar()
        with Container(id="chat"):
            yield ChatHeader(self.client.host, self.client.port)
            yield VimVerticalScroll(id="messages")
            with HorizontalGroup(id="input-wrapper"):
                yield Label("[bold]>[/]", id="prompt-char")
                yield TextInput(compact=True)

    async def action_send(self):
        textarea = self.query_one(TextInput)
        if not textarea.text:
            return
        await self.client.send_text(textarea.text)
        textarea.clear()

    async def action_quit(self):
        await self.app.cleanup()
        self.app.exit()

    def action_show_help(self):
        self.app.push_screen(HelpScreen())

    def action_open_menu(self):
        self.app.push_screen(MenuScreen())

    async def on_mount(self):
        self.app.message_worker = self.run_worker(self.update_messages)
        self.query_one(TextInput).focus()
        await self.client.request_nicknames()

    async def update_messages(self):
        while True:
            try:
                msg = await self.client.receive_message()
            except SackClientServerError:
                self.post_message(self.ServerDown())
                break
            if msg is None:
                continue
            self.post_message(self.MessageReceived(msg))

    @on(ServerDown)
    async def on_server_down(self):
        await self.app.cleanup()
        self.app.back_to_first_screen()
        self.app.push_screen(ServerDownScreen())

    @on(MessageReceived)
    def on_message_received(self, event: MessageReceived):
        msg = event.msg
        messages = self.query_one("#messages", VimVerticalScroll)
        users = self.query_one("#sidebar-users", Container)
        if msg.type == "CONNECT":
            if msg.username == self.client.username:
                return
            notif = Label(f"{msg.username} joined", classes="notification")
            user = get_sidebar_user(msg.username, self.colors_manager.get(msg.username))
            messages.mount(notif)
            users.mount(user)
            notif.scroll_visible()
        if msg.type == "DISCONNECT":
            notif = Label(f"{msg.username} disconnected", classes="notification")
            messages.mount(notif)
            notif.scroll_visible()
            color = self.colors_manager.get(msg.username)
            with suppress(Exception):
                user = users.query_one(f"#{get_id_from_color(color)}")
                user.remove()
        if msg.type == "TEXT":
            assert msg.text
            if msg.username == self.username:
                orientation = "right"
                color = None
            else:
                orientation = "left"
                color = self.colors_manager.get(msg.username)
            new_msg = ChatMessage(orientation=orientation, msg=msg.text, author=msg.username, color=color)
            messages.mount(new_msg)
            new_msg.scroll_visible()
        if msg.type == "GETNICKNAMES":
            assert msg.text
            for u in msg.text.split("\n"):
                users.mount(get_sidebar_user(u, self.colors_manager.get(u), is_you=u == self.client.username))


class HelpScreen(ModalScreen):
    BINDINGS = [
        Binding("escape", "app.pop_screen"),
        Binding("l", "tab_next"),
        Binding("h", "tab_previous"),
        Binding("j", "scroll_down"),
        Binding("k", "scroll_up"),
    ]

    def compose(self) -> ComposeResult:
        with Container(classes="help"):
            with HorizontalGroup(id="help-buttons"):
                yield HelpTab("Welcome", "welcome")
                yield HelpTab("Create server", "server")
                yield HelpTab("Join server", "join")
                yield HelpTab("Chat", "chat_")
            with ContentSwitcher(initial="welcome"):
                with VerticalScroll(id="welcome", can_focus=False):
                    yield Static(WELCOME_HELP, classes="help-paragraph")
                    with Center():
                        yield Label("Keybindings", classes="help-subtitle")
                    for key, desc in WELCOME_KB:
                        yield HelpKeybinding(key, desc)
                with VerticalScroll(id="server", can_focus=False):
                    yield Static(SERVER_HELP, classes="help-paragraph")
                    with Center():
                        yield Label("Keybindings", classes="help-subtitle")
                    for key, desc in FORMS_KB:
                        yield HelpKeybinding(key, desc)
                with VerticalScroll(id="join", can_focus=False):
                    yield Static(JOIN_HELP, classes="help-paragraph")
                    with Center():
                        yield Label("Keybindings", classes="help-subtitle")
                    for key, desc in FORMS_KB:
                        yield HelpKeybinding(key, desc)
                with VerticalScroll(id="chat_", can_focus=False):
                    yield Static(CHAT_HELP, classes="help-paragraph")
                    with Center():
                        yield Label("Keybindings", classes="help-subtitle")
                    for key, desc in CHAT_KB:
                        yield HelpKeybinding(key, desc)
            yield Label(make_keybinding_text(HELP_KB), id="help-footer")

    def action_tab_next(self) -> None:
        self.focus_next()
        self.change_tab_from_focused()

    def action_tab_previous(self) -> None:
        self.focus_previous()
        self.change_tab_from_focused()

    def change_tab_from_focused(self):
        focused = self.focused
        if isinstance(focused, Button) and focused.id:
            self.query_one(ContentSwitcher).current = focused.id

    def action_scroll_up(self) -> None:
        self.get_current_content().scroll_up()

    def action_scroll_down(self) -> None:
        self.get_current_content().scroll_down()

    def get_current_content(self) -> VerticalScroll:
        current_id = self.query_one(ContentSwitcher).current
        scrolls = self.query(VerticalScroll)
        for s in scrolls:
            if s.id == current_id:
                return s
        raise AssertionError("")


class ThemeChangeScreen(ModalScreen):
    BINDINGS = [
        Binding("escape", "app.pop_screen"),
        Binding("enter", "app.pop_screen", priority=True),
        Binding("down, tab, j", "focus_next", priority=True),
        Binding("up, shift+tab, k", "focus_previous", priority=True),
    ]

    def action_focus_next(self) -> None:
        self.focus_next()
        self.change_theme_from_focused()

    def action_focus_previous(self) -> None:
        self.focus_previous()
        self.change_theme_from_focused()

    def change_theme_from_focused(self):
        focused = self.focused
        if isinstance(focused, Button) and focused.id:
            assert focused.id in self.app.available_themes
            self.app.theme = focused.id

    def compose(self) -> ComposeResult:
        with Container(classes="modal"):
            with Center():
                yield Label("Change theme", classes="modal-title")
            for theme in self.app.available_themes:
                yield Option(theme, theme)


class MenuScreen(ModalScreen):
    app: "SackApp"

    BINDINGS = COMMON_BINDINGS

    def compose(self) -> ComposeResult:
        with Options(classes="modal"):
            with Center():
                yield Label("Menu", classes="modal-title")
            yield Option("Exit app", "exit")
            yield Option("Exit to menu", "exit_to_menu")
            yield Option("Help", "show_help")

    async def on_button_pressed(self, e: Button.Pressed):
        match e.button.id:
            case "exit":
                await self.app.cleanup()
                self.app.exit()
            case "exit_to_menu":
                await self.app.cleanup()
                self.app.back_to_first_screen()
            case "show_help":
                self.app.pop_screen()
                self.app.push_screen(HelpScreen())


class ServerDownScreen(ModalScreen):
    BINDINGS = [Binding("enter, escape", "app.pop_screen", priority=True)]

    def compose(self) -> ComposeResult:
        with Container(classes="modal"):
            with Center():
                yield Label("Connection to the server was lost", classes="modal-title")


class AboutScreen(Screen):
    app: "SackApp"

    BINDINGS = [Binding("escape", "app.pop_screen")]

    def compose(self) -> ComposeResult:
        yield from self.app.get_header()
        with Container(classes="container"):
            yield Label(SACK_ABOUT)
        with Container(id="footer-container"):
            yield Label(make_keybinding_text(ABOUT_KB), classes="container")
